/*
 * $Id: brctl.c,v 1.4 2000/01/17 23:10:15 root Exp $
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <linux/br.h>
#include <asm/param.h>
#include "brctl.h"

#define TOKEN_NONE 0
#define TOKEN_ADDBR 1
#define TOKEN_DELBR 2
#define TOKEN_ADDIF 3
#define TOKEN_DELIF 4

static int br_fd;
static int sock_fd;

void dump_bridge_id(unsigned char *x)
{
	printf("%.2x%.2x.%.2x%.2x%.2x%.2x%.2x%.2x", x[0], x[1], x[2], x[3],
	       x[4], x[5], x[6], x[7]);
}

void get_fds()
{
	if ((br_fd = open("/dev/brctl", O_RDWR)) < 0) {
		fprintf(stderr, "bridging was not compiled into your kernel or /dev/brctl\n");
		fprintf(stderr, "could not be opened.\n");
		exit(-1);
	}

	if ((sock_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		perror("socket");
		exit(-1);
	}
}

void br_addbr(char *br)
{
	char _br[32];

	memcpy(_br, br, 32);
	if (ioctl(br_fd, BRCTL_ADD_BRIDGE, _br) > 0)
		return;

	switch (errno) {
	case 0:
		break;

	case EEXIST:
		fprintf(stderr, "bridge %s already exists!\n", br);
		break;

	default:
		perror("ioctl");
		exit(-1);
	}
}

void br_delbr(char *br)
{
	char _br[32];

	memcpy(_br, br, 32);
	if (ioctl(br_fd, BRCTL_DEL_BRIDGE, _br) > 0)
		return;

	switch (errno) {
	case 0:
		break;

	case ENXIO:
		fprintf(stderr, "bridge %s doesn't exist!\n", br);
		break;

	case EBUSY:
		fprintf(stderr, "bridge %s is still up!\n", br);
		break;

	default:
		perror("ioctl");
		exit(-1);
	}
}

void br_addif(char *br, char *iface, int ifindex)
{
	struct ifreq ifr;
	unsigned long *i;

	memcpy(ifr.ifr_name, br, 16);
	i = (unsigned long *)(&ifr.ifr_data);
	i[0] = BRCTL_ADD_IF;
	i[1] = ifindex;
	if (ioctl(sock_fd, SIOCDEVPRIVATE, &ifr) > 0)
		return;

	switch (errno) {
	case 0:
		break;

	case EBUSY:
		fprintf(stderr, "device %s already enslaved!\n", iface);
		break;

	default:
		perror("ioctl");
		exit(-1);
	}
}

void br_delif(char *br, char *iface, int ifindex)
{
	struct ifreq ifr;
	unsigned long *i;

	memcpy(ifr.ifr_name, br, 16);
	i = (unsigned long *)(&ifr.ifr_data);
	i[0] = BRCTL_DEL_IF;
	i[1] = ifindex;
	if (ioctl(sock_fd, SIOCDEVPRIVATE, &ifr) > 0)
		return;

	switch (errno) {
	case 0:
		break;

	case EINVAL:
		fprintf(stderr, "device %s is not a slave of %s!\n", iface, br);
		break;

	default:
		perror("ioctl");
		exit(-1);
	}
}

void br_dumpbridges()
{
	int i;
	int ifindices[32];
	int num;

	ifindices[0] = 32;
	num = ioctl(br_fd, BRCTL_GET_BRIDGES, ifindices);
	if (num < 0) {
		perror("ioctl");
		return;
	}
	for (i=0;i<num;i++)
		printf("%i\n", ifindices[i]);
}

void br_setstpstate(char *br, int state)
{
	unsigned long *i;
	struct ifreq ifr;

	memcpy(ifr.ifr_name, br, 16);
	i = (unsigned long *)(&ifr.ifr_data);
	i[0] = BRCTL_SET_BRIDGE_STP_STATE;
	i[1] = state;
	if (ioctl(sock_fd, SIOCDEVPRIVATE, &ifr) < 0) {
		perror("setstpstate");
		return;
	}
}

void br_dumpindices(char *br)
{
	unsigned long *i;
	struct ifreq ifr;
	int indices[256];
	int j;

	memcpy(ifr.ifr_name, br, 16);
	i = (unsigned long *)(&ifr.ifr_data);
	i[0] = BRCTL_GET_PORT_LIST;
	i[1] = (unsigned long)indices;
	if (ioctl(sock_fd, SIOCDEVPRIVATE, &ifr) < 0) {
		perror("dumpindices");
		return;
	}

	for (j=0;j<256;j++)
		if (indices[j])
			printf("%i %i\n", j, indices[j]);
}

char *state[5] = {"disabled", "listening", "learning", "forwarding", "blocking"};

void br_dumpportinfo(char *br, int port)
{
	unsigned long *i;
	struct ifreq ifr;
	struct port_info p;

	memcpy(ifr.ifr_name, br, 16);
	i = (unsigned long *)(&ifr.ifr_data);
	i[0] = BRCTL_GET_PORT_INFO;
	i[1] = (unsigned long)&p;
	i[2] = port;
	if (ioctl(sock_fd, SIOCDEVPRIVATE, &ifr) < 0)
		return;

	printf("port id\t\t\t%.2x\n", p.port_id);
	printf("state\t\t\t%s\t\tflags\t", state[p.state]);
	if (p.top_change_ack)
		printf("TOPOLOGY_CHANGE_ACK ");
	if (p.config_pending)
		printf("CONFIG_PENDING ");
	printf("\ndesignated root\t\t");
	dump_bridge_id((unsigned char *)&p.designated_root);
	printf("\tdesignated port\t\t%.2x\n", p.designated_port);
	printf("designated bridge\t");
	dump_bridge_id((unsigned char *)&p.designated_bridge);
	printf("\nmessage age\t\t%4i\t\t\t", p.message_age_timer_value/HZ);
	printf("path cost\t\t%4i\n", p.path_cost);
	printf("forward delay\t\t%4i\t\t\t", p.forward_delay_timer_value/HZ);
	printf("designated cost\t\t%4i\n", p.designated_cost);
	printf("hold timer\t\t%4i\n", p.hold_timer_value/HZ);
	printf("\n");
}

void br_dumpinfo(char *br)
{
	struct bridge_info b;
	unsigned long *i;
	struct ifreq ifr;
	int j;

	memcpy(ifr.ifr_name, br, 16);
	i = (unsigned long *)(&ifr.ifr_data);
	i[0] = BRCTL_GET_BRIDGE_INFO;
	i[1] = (unsigned long)&b;
	if (ioctl(sock_fd, SIOCDEVPRIVATE, &ifr) < 0) {
		perror("dumpinfo");
		return;
	}

	if (!b.stp_enabled) {
		printf("STP disabled\n");
		return;
	}

	printf("bridge\t\t\t%s\n", br);
	printf("bridge_id\t\t");
	dump_bridge_id((unsigned char *)&b.bridge_id);
	printf("\ndesignated root\t\t");
	dump_bridge_id((unsigned char *)&b.designated_root);
	printf("\nflags\t\t\t");
	if (b.topology_change)
		printf("TOPOLOGY_CHANGE ");
	if (b.topology_change_detected)
		printf("TOPOLOGY_CHANGE_DETECTED ");
	printf("\nroot port\t\t%3i\t\tpath cost\t\t%3i\n",
	       b.root_port, b.root_path_cost);
	printf("max age\t\t\t%3i\t\tbridge max age\t\t%3i\n",
	       b.max_age/HZ, b.bridge_max_age/HZ);
	printf("hello time\t\t%3i\t\tbridge hello time\t%3i\n",
	       b.hello_time/HZ, b.bridge_hello_time/HZ);
	printf("forward delay\t\t%3i\t\tbridge forward delay\t%3i\n",
	       b.forward_delay/HZ, b.bridge_forward_delay/HZ);
	printf("topology change time\t%3i\t\thold time\t\t%3i\n",
	       b.topology_change_time/HZ, b.hold_time/HZ);
	printf("\n");
	printf("\n");

	for (j=1;j<256;j++)
		br_dumpportinfo(br, j);
}

int token(char *token)
{
	if (!strcasecmp(token, "addbr")) return TOKEN_ADDBR;
	if (!strcasecmp(token, "delbr")) return TOKEN_DELBR;
	if (!strcasecmp(token, "addif")) return TOKEN_ADDIF;
	if (!strcasecmp(token, "delif")) return TOKEN_DELIF;

	return TOKEN_NONE;
}

void help()
{
	fprintf(stderr, "syntax:\n");
	fprintf(stderr, "brctl addbr <bridge name>\t\t\tadd bridge\n");
	fprintf(stderr, "brctl delbr <bridge name>\t\t\tdelete bridge\n");
	fprintf(stderr, "brctl addif <bridge name> <interface name>\tadd interface to bridge\n");
	fprintf(stderr, "brctl delif <bridge name> <interface name>\tdelete interface from bridge\n");
}

int main(int argc, char *argv[])
{
	int needhelp = 1;
	int index;

	if (argc == 1) {
		help();
		return 1;
	}

	get_fds();
	br_dumpbridges();
	br_dumpportinfo("br0", 1);
	br_setstpstate("br0", 1);

	index = 1;
	while (index < argc) {
		char *br;
		char *iface;
		int t;
		int ifindex;

		t = token(argv[index]);
		if (t == TOKEN_NONE) {
			needhelp = 1;
			break;
		}

		br = argv[++index];
		iface = argv[++index];
		needhelp = 0;

		switch (t) {
		case TOKEN_ADDBR:
			br_addbr(br);
			index += 2;
			break;

		case TOKEN_DELBR:
			br_delbr(br);
			index += 2;
			break;

		case TOKEN_ADDIF:
		case TOKEN_DELIF:
			index++;

			if (!(ifindex = if_nametoindex(iface))) {
				fprintf(stderr, "interface %s not found!\n", iface);
				break;
			}

			if (t == TOKEN_ADDIF)
				br_addif(br, iface, ifindex);
			else
				br_delif(br, iface, ifindex);

			break;
		}

		br_dumpindices(br);
		br_dumpinfo(br);
	}

	if (needhelp)
		help();

	return 0;
}
